/* Arduino sketch_String_Decoder_Rev_10.  Jon Titus.
 * Arduino 1.8.5.
 * This program demonstrates how to parse a JSON payload in function JSON_Parse.  It takes a String name and returns  
 *    an error code(90-99) or number of name-value pairs in a JSON payload if no errors.
 * Serial port runs at 115,200 bits/sec, 8-N-1 format.
 * It provides the JSON-parsing code in a function.  Results are in three arrays.
 * Uses the MQTT topic esp/test.  You may change this.
 * 05-01-2019 at 1046H MDT.  Arduino sketch_String_Decoder_Rev_10.  Jon Titus.
 * MIT Copyright license, 2019, Jon Titus, Herriman, UT.
 * 
 */
  // ----- START of PROGRAM PACKAGE-----
  
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
 
const char* ssid =         "         ";               //Titus home WiFi network name
const char* password =     "         ";               //Password
const char* mqttServer =   "???.cloudmqtt.com";       //Server name assigned by mqttcloud Web site registration
const int   mqttPort =      aValue;                    //Port number assigned
const char* mqttUser =     "        ";                //User name assigned
const char* mqttPassword = "            ";            //mqttcloud assigned password
String testString = "";                               //String for complete message
const float bigNumber = 0xFFFFFFFFFF;                 //Very large number (trillions!) 
WiFiClient espClient;                                 //Initialize ESP client
PubSubClient client(espClient);                       //Constructor for "client" of object PubSubClient

//********************************************  
// ----- DEFINE/DECLARE GLOBALS -----
  char leftBrace =     0x7B;     //ASCII code for {
  char rightBrace =    0x7D;     //ASCII code for }
  char qMark =         0x3F;     //ASCII code for ?
  char plusSign =      0x2B;     //ASCII code for +
  char minusSign =     0x2D;     //ASCII code for -
  char charBelowZero = 0x2F;     //ASCII for "0" minus 1 (0x30 - 1 = 0x2F)
  char charAboveNine = 0x3A;     //ASCII for "9" plus 1 (0x39 + 1 = 0x3A)
  char quoteMark =     0x22;     //ASCII for "
  char commaMark =     0x2C;     //ASCII for comma
  
//Declare global arrays to save message names, text, and numeric info
  #define arraySize 5                 //Change this value as needed
  String JName[arraySize];            //Name, always starts with text
  String JValue[arraySize];           //Message text, if message starts with a character,
                                      //  otherwise, empty.
  float  JNumber[arraySize];          //Number, if message starts with 0-9, +, or -,
                                      //  otherwise bigNumber (0xFFFFFFFFFF).
//********************************************

// ----- START CALLBACK ROUTINE -----
// This function runs when the client.loop() function detects receipt of a new message
void callback(char* topic, unsigned char* payload, unsigned int length)     //Get topic, payload, and length
{
  String testString = "";                             //Initialize an empty string
  Serial.print("\n\rMessage arrived in topic: ");     //Message arrived, so print topic
  Serial.println(topic);
  Serial.print("Payload: ");                          //Print "Payload: "
  Serial.println((char*)payload);
  Serial.println("-----------------------");
  
  //Known-good String for testing if you need it
  //String testString = "{\"SensorType\":\"Temperature\",\"Value\":10,\"Testing\":\"ON\",\"Another value\":987.437}";

  int returnCode = JSON_Parse((char*)payload);        //Parse the TestString and print values
  switch (returnCode)                                 //Test return value for any error codes
   {
    case 97:
      Serial.println("Error in Value code.");
      break;
    case 98:
      Serial.println("No colon found.");
      break;
    case 99:
      Serial.println("Invalid message.");  
   }      
   if (returnCode < 90)                           //No error codes, test returned number of message pairs
   {                                              //Less than 95 (change as needed)
    Serial.println("Parsing OK");                 //Print message and print pairs of values
    Serial.println("-----------------------------");
    for (int arrayCounter = 0; arrayCounter < returnCode; arrayCounter++)
      {
        //After you have values, change the Serial.print statements to whatever you want for a name-value pair.
        //The following statements display the results of JSON parsing.
        Serial.print("Name-Value pair#: ");                    //Print the message Name                  
        Serial.println(arrayCounter);
        Serial.print("JName: ");
        Serial.println(JName[arrayCounter]);
        if (JValue[arrayCounter] != "")                 //If value holds something               
        {
          Serial.print("JValue: ");                     //Print the value
          Serial.println(JValue[arrayCounter]);
        }
        if (JNumber[arrayCounter] != bigNumber)         //If the number not equal to 0xFFFFFFFFFF
        {
          Serial.print("JNumber: ");                    //Print the number
          Serial.println(JNumber[arrayCounter]);
        }
        Serial.println("-----------------------------");
        //Serial.println(" ");
      }
   }
}
// ----- END CALLBACK ROUTINE -----
//********************************************

// ----- START SETUP OPERATIONS -----
void setup()
 {   
   Serial.begin(115200);                            //Serial set at 115,200 bits/sec.  Default: 8N1 format
   Serial.println();                                //Clear UART buffer of any leftover "junk" bytes
   WiFi.begin(ssid, password);                      //Start local WiFi connection
  while (WiFi.status() != WL_CONNECTED)             //Loop here until local WiFi connects
  {                                                 //Eight possible reply messages
    delay(500);                                     //Short delay, 500 msec.
    Serial.println("Connecting to WiFi..");         //Print this in the loop
  }
  Serial.println("Connected to the WiFi network");  //Connected to local WiFi
  client.setServer(mqttServer, mqttPort);           //Set MQTT-server connection
  client.setCallback(callback);                     //Declare callback for when msg received
  while (!client.connected())
  {
    Serial.println("Connecting to MQTT...");                      //Print this info
    if (client.connect("ESP32Client", mqttUser, mqttPassword ))   //Wait for "true" response
    {
       Serial.println("connected");                 //OK, you connected...  
     }
     else                                           //If not true, print diagnostic info
     {
       Serial.print("Failed with state ");          //Get diagnostic info
       Serial.print(client.state());                //Get error data and print it
       delay(2000);                                 //Wait 2 seconds, then try again
    }
  }
  client.subscribe("esp/test");                     //Subscribe this device to esp/test topic. Change to any topic name you wish.
}                                                   //Ensure your publishing devices use the same topic name!!!

// ----- END SETUP OPERATIONS -----
//********************************************
// ----- START VOID LOOP: Continue to test for and accept messages as they arrive -----
void loop()
{ 
    client.loop();
}
// ----- END VOID: LOOP -----
//******************************************** 
// ----- JSON PARSING FUNCTION -----
//For reference on Arduino C++ String objects: https://www.arduino.cc/reference/en/language/variables/data-types/stringobject/.

int JSON_Parse(String Jstring)  //Enter this function with a string object
{
  // ----- LOCAL VARIABLES -----
  int Jlength;                  //Length of entire JSON string
  int quotestart;               //pos'n of first quote in a name or value
  int quoteend;                 //pos'n of last quote in a name or value
  int numberEnd;                //String pos'n of comma after numeric value
  int index;                    //Loop counter
  int dataIndex;                //One less than number of NAME/VALUE pairs 
  char signTest;                //Result from test of 1st char in a numeric value
  
  Jlength = (Jstring.length());                     //Get length of string; printing character. Note: \" = one char
  int numbOfCommas = 0;                             //Reset number of commas from previous JSON object
  for (index = 0; index < Jlength; index++)         //Loop to count commas; # of commas is one less than
  {                                                 //  number of message/value pairs
     if ((char)Jstring.charAt(index) == 0x2C)       //Find a comma in the JSON payload
     {
        numbOfCommas += 1;                          //Accumulate number of commas
     }
  }
  if(((char)Jstring.charAt(0) != leftBrace) || (char)Jstring.charAt(Jlength - 1) != rightBrace)    //Test for "{" and "}" in JSON object
  {                                                                                                // at start and end.
    goto error99;                                                   //Missing bracket triggers error.
  }
  for (dataIndex = 0; dataIndex < (numbOfCommas+1); dataIndex++)    //Main message-decoding steps in this section
  {                                                                 //Get JSON message's NAME between quotes
    quotestart = Jstring.indexOf("\"");                             //Find quote at start of section & save location
    Jstring.setCharAt(quotestart, qMark);                           //Replace quote mark with a ?-mark, just a place holder
    quoteend = Jstring.indexOf("\"");                               //Find quote at end of section & save location
    Jstring.setCharAt(quoteend, qMark);                             //Replace quote mark with a ?-mark, just a place holder
    JName[dataIndex]= Jstring.substring(quotestart+1, quoteend);    //Start at first char after first quote and go to last quote
    
    //Test for a colon that separates NAME and VALUE information
    if((char)Jstring.charAt(quoteend+1) != 0x3A)                //Look at character after the last quote in text.  Is is colon? ":" 
      {
      goto error98;                                             //No colon found, error message
      }
    Jstring.setCharAt(quoteend + 1, qMark);                     //Yes, colon found. Replace colon with a ? mark, just placeholder.
    if ((char)Jstring.charAt(quoteend + 2) == quoteMark)        //Is next char a quote mark?
    {                                                           //Yes, it is a quote mark, continue...
        quotestart = Jstring.indexOf("\"");                     //Find first quote in VALUE message 
        Jstring.setCharAt(quotestart, qMark);                   //Replace quote mark with a ?-mark as a placeholder.
        quoteend = Jstring.indexOf("\"");                       //Find location of ending quote mark                               
        Jstring.setCharAt(quoteend, qMark);                     //Replace quote mark with a ?-mark as a placeholder.
        if((char)Jstring.charAt(quoteend + 1) == commaMark)     //If there is a comma after the quote mark, replace it with a ? mark.
          {
            Jstring.setCharAt(quoteend + 1, qMark);
          }           
        JValue[dataIndex] = Jstring.substring(quotestart+1, quoteend);        //Start after first quote and get string up to last quote.
        JNumber[dataIndex] = bigNumber;                                       //Put bigNumber here as signal to NOT use it.
    }
    else                                                                      //No quote after the colon, so we have a numeric value
    {
        signTest = Jstring.charAt(quoteend + 2);                              //Get first character after the colon.  What is it?       
                                                                              //Is 1st char == +, -, or a number..?
        if((signTest == plusSign) || (signTest == minusSign) || ((signTest > charBelowZero) && (signTest < charAboveNine)))
          {                                                                   //Yes
            if (Jstring.indexOf(",") != -1)                                   //Value can end with a comma or a rightBrace
              {                                                               //Test for comma
                numberEnd = Jstring.indexOf(",");                             //If a comma at the end of numeric value save its index
                Jstring.setCharAt(numberEnd, qMark);                          //Replace comma with "?" as a place holder
              }
              else if ((Jstring.indexOf("}")) != -1)                          //Test for right brace at end of value
              {
                //Serial.println("A brace ends it");                          //For testing
                numberEnd = Jstring.indexOf("}");                             //if number ends with a }, save its string index 
              }       
        JValue[dataIndex] = Jstring.substring(quoteend+2, numberEnd);         //Get numeric string, incl sign and decimal point
        JNumber[dataIndex] = JValue[dataIndex].toFloat();                     //Convert value in string to float
        JValue[dataIndex] = "";                                               //Clear anything in the text VALUE string.
      }
      else
      {
        goto error97;
      }
    }
  }
  return(numbOfCommas + 1);      //Everything parsed properly, return number of message pairs
  
 // Error codes.
  error99:
    return(99); 
  error98:
    return(98); 
  error97:
    return(97);
}
//********************************************
// ----- END JSON PARSING FUNCTION -----
//----- END -----
