/*
 * sketch_TempHumid_JSON_Rev_2
 * Some software taken from TECHTUTORIALSX Web site.  Thank you Rui Sanchez.
 * Tutorial "ESP8266: Connecting to MQTT broker," dated April 9, 2017.
 * 
 * Modifications and comments added by Jon Titus, Herriman, UT 03-08-2019
 * Now includes sensor reading, creation of JSON message and Internet
 * WiFi connection.  Revised to add "Publish Success" and "Publish Failure" messages for MQTT reply
 * included callback declaration and function.  
 * Refer to JSON Style Guide: https://google.github.io/styleguide/jsoncstyleguide.xml
 * 
 * Added code to create a JSON message.  Received properly at the MQTTLens (Google app) window.

 * Tested: 03-10-2019 at 1150H MDT.  File: sketch_TempHumid_JSON_Rev_2. Jon Titus
 * IT WORKS
 * Ensure you have Adafruit_Sensor.h file in the DHT library folder!
 */
// ----- START OF PROGRAM ----- 
#include <ESP8266WiFi.h>                              //Include Arduino ESP8266 library
#include <PubSubClient.h>                             //Include MQTT library                              
#include "DHT.h"                                      //Include the humidity & temp sensor library

#define DHTTYPE DHT22                                 // DHT 22  (AM2302), AM2321
const int DHTPin = 5;                                 //Pin D1 is gpio5 for DHT Sensor
DHT dht(DHTPin, DHTTYPE);  

const char* ssid =         "         ";               //Titus home WiFi network name
const char* password =     "         ";               //Password
const char* mqttServer =   "???.cloudmqtt.com";       //Server name assigned by mqttcloud Web site registration
const int   mqttPort =      aValue;                   //Port number assigned
const char* mqttUser =     "        ";                //User name assigned
const char* mqttPassword = "            ";            //mqttcloud assigned password
 
WiFiClient espClient;                                 //Initialize espClient
PubSubClient client(espClient);                       //Constructor for "client" of object PubSubClient

// ----- START OF SETUP OPERATIONS -----
void setup()
{ 
  Serial.begin(115200);                               //Bit rate for Arduino terminal
  WiFi.begin(ssid, password);                         //Start local WiFi connection
  dht.begin();                                        //Start DHT humidity & temp sensor
  
  while (WiFi.status() != WL_CONNECTED)               //Loop here until WiFi local connects
  {                                                   // eight possible reply messages
    delay(500);                                       //Wait 500 msec between tests
    Serial.println("Connecting to WiFi..");           //Print this message while waiting
  }
  Serial.println("Connected to the WiFi network");    //Print this message when WiFi connects
  client.setServer(mqttServer, mqttPort);             //Set MQTT server information
  client.setCallback(callback);                       //Pointer for callback function, active when msg received
                                                      
  while (!client.connected())                         //Wait while "client" not connected (T or F)
  {
    Serial.println("Connecting to MQTT...");          //Print message          
    if (client.connect("ESP8266Client", mqttUser, mqttPassword ))   //if client connects to MQTT server, then
    {
      Serial.println("connected to MQTT");                          //  print this message  
    }
    else                                              //If client does not connect
    {
      Serial.print("Failed with state ");             //If connection to server failed
      Serial.print(client.state());                   //Print client state
      delay(2000);                                    //Delay 2 sec.
    }
  }                                                         
  client.subscribe("esp/test");                       //Also subscribe to topic esp/test
}
// ----- END OF SETUP OPERATIONS -----

// ----- START OF CALLBACK FUNCTION -----
void callback(char* topic, byte* payload, unsigned int length)  //Message received from server: topic, payload (data)
{                                                               // length of payload (number of bytes)
  Serial.print("Message arrived in topic: ");
  Serial.println(topic);                                        //Print topic of message, esp/test, for example
  Serial.print("Message :"); 
  for (int i = 0; i < length; i++)                              //Print payload, char by char
  {
    Serial.print((char)payload[i]);
  }
  Serial.println();                                             //Finished printing results
  Serial.println("-----------------------");
}
// ----- END OF CALLBACK FUNCTION -----

// ----- START OF dataUpdate FUNCTION -----

void dataUpdate()                                         //Update the temp (F & C) and humidity readings
{
  float h = dht.readHumidity();           
  float c = dht.readTemperature();
  float f = dht.readTemperature(true);           
  if (isnan(h) || isnan(c) || isnan(f))                   // Check if any reads failed and exit early (to try again).
  {
    Serial.println("Failed to read from DHT sensor!");    //Error message to local terminal         
  }            
  String humidValue = String(h);                          //Convert float data to string
  String celsiValue = String(c);
  String fahreValue = String (f);
  String payloadout = "{";                                //Create JSON payloadout to send to MQTT server
  payloadout += "\"Humidity\":";                          //Refer to: JSON Style Guide
  payloadout += humidValue;
  payloadout += ",";
  payloadout += "\"TempC\":";
  payloadout += celsiValue;
  payloadout += ",";
  payloadout += "\"TempF\":";
  payloadout += fahreValue;
  payloadout += "}";
  Serial.println(payloadout);      
  char attributes[200];
  payloadout.toCharArray(attributes, 200);            //Create character array
  client.publish("esp/test", attributes);             //Publish the name and value pairs to server                                                     // in JSON format              
}
// ----- END OF TICKER FUNCTION -----

// ----- START OF MAIN LOOP -----
void loop()
{
  for(int loopcounter = 0; loopcounter < 1000000; loopcounter++)
  {
    client.loop();                                      //This loop continues to test for new
  }
  dataUpdate();
}                                                     // messages from subscribed topic.
// ----- END OF MAIN LOOP -----                       // new message triggers callback function
// ----- END OF PROGRAM -----
